.doc-label-async code {
    color: #fc7676;
    font-style: italic;
}

.md-typeset a.toc-permalink {
    font-size: 0.6em;
    color: #97bbff86;
    padding-left: 3px;
}

:root  > * {
    --md-primary-fg-color:        #002c5b;
  }

.md-header__button.md-logo img, .md-header__button.md-logo svg {
    fill: #00c1bc;
}
/*
 Heading Styles */

.md-typeset h5 .doc-object-name {
    text-transform: none !important;
}

/* .md-typeset h3.doc-heading {
    font-size:1.2em;
}

.md-typeset h4.doc-heading {
    font-size:1.17em;
}

.md-typeset h5.doc-heading {
    font-size:1.0em;
} */

.doc-labels .doc-label-instance-attribute code {
    color: #70d381;
}

.doc-labels .doc-label-property code {
    color: #d39977;
}