"""Data models for VeSync API requests and responses.

Models should follow the naming convention of Request/Response + API Name.
Internal models can have any descriptive name.

The `base_models` module contains the base classes for the VeSync API models, while the
models for each device type and other API calls are stored in their respective modules.

The [bypassv2_models][pyvesync.models.bypass_models] module contains the models for
the common bypassV2 API calls.
"""
